# Generated by Django 3.1.3 on 2020-12-08 17:53

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0005_hypothesis_and_created_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='abtest',
            name='sample_size',
            field=models.PositiveIntegerField(validators=[django.core.validators.MinValueValidator(1)]),
        ),
    ]



import datetime
import json

from django import forms
from django.core.exceptions import PermissionDenied
from django.db.models import Sum, Q, F
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.utils import formats, timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext as _, gettext_lazy
from django.views.decorators.csrf import csrf_exempt
import django_filters
from django_filters.constants import EMPTY_VALUES
from rest_framework import status
from rest_framework.decorators import (
    api_view,
    authentication_classes,
    permission_classes,
)
from rest_framework.response import Response
from wagtail.admin import messages, panels
from wagtail.admin.action_menu import ActionMenuItem
from wagtail.admin.filters import DateRangePickerWidget, WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.models import Page, PAGE_MODEL_CLASSES

from .models import AbTest
from .events import get_event_types


class CreateAbTestForm(forms.ModelForm):
    goal_event = forms.ChoiceField(choices=[])
    hypothesis = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["goal_event"].choices = [
            (slug, goal.name) for slug, goal in get_event_types().items()
        ]

    def save(self, page, variant_revision, user):
        ab_test = super().save(commit=False)
        ab_test.page = page
        ab_test.variant_revision = variant_revision
        ab_test.created_by = user
        ab_test.save()
        return ab_test

    class Meta:
        model = AbTest
        fields = ["name", "hypothesis", "goal_event", "goal_page", "sample_size"]

    panels = [
        panels.MultiFieldPanel(
            [
                panels.FieldPanel("name"),
                panels.FieldPanel("hypothesis"),
            ],
            heading=_("Enter test details"),
        ),
        panels.MultiFieldPanel(
            [
                # A dummy help panel to mount the react component:
                panels.HelpPanel(attrs={"data-component": "goal-selector"}),
            ],
            heading=_("Choose a goal"),
        ),
        panels.MultiFieldPanel(
            [
                panels.FieldPanel("sample_size"),
                panels.HelpPanel(
                    _(
                        "Need help calculating sample size for A/B tests? "
                        'Try <a href="https://www.optimizely.com/uk/sample-size-calculator/" '
                        'target="_blank">Optimisely</a>, '
                        '<a href="https://www.evanmiller.org/ab-testing/sample-size.html" '
                        'target="_blank">Evan Miller</a>, or '
                        '<a href="https://www.abtasty.com/sample-size-calculator/" '
                        'target="_blank">AB Tasty</a>.'
                    )
                ),
            ],
            heading=_("Sample size"),
        ),
    ]


def add_ab_test_checks(request, page):
    # User must have permission to edit the page
    page_perms = page.permissions_for_user(request.user)
    if not page_perms.can_edit():
        raise PermissionDenied

    # User must have permission to add A/B tests
    if not request.user.has_perm("wagtail_ab_testing.add_abtest"):
        raise PermissionDenied

    # Page must not already be running an A/B test
    if AbTest.objects.get_current_for_page(page=page):
        messages.error(request, _("This page already has a running A/B test"))

        return redirect("wagtailadmin_pages:edit", page.id)

    # Page must be published and have a draft revision
    if not page.live or not page.has_unpublished_changes:
        messages.error(
            request,
            _("To run an A/B test on this page, it must be live with draft changes."),
        )

        return redirect("wagtailadmin_pages:edit", page.id)


def add_compare(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    # Run some checks
    response = add_ab_test_checks(request, page)
    if response:
        return response

    latest_revision_as_page = page.get_latest_revision_as_object()

    comparison = (
        page.get_edit_handler()
        .get_bound_panel(page, request=request, form=None)
        .get_comparison()
    )

    comparison = [comp(page, latest_revision_as_page) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return render(
        request,
        "wagtail_ab_testing/add_compare.html",
        {
            "page": page,
            "latest_revision_as_page": latest_revision_as_page,
            "comparison": comparison,
            "differences": any(comp.has_changed() for comp in comparison),
        },
    )


def add_form(request, page_id):
    page = get_object_or_404(Page, id=page_id)

    # Run some checks
    response = add_ab_test_checks(request, page)
    if response:
        return response

    if request.method == "POST":
        form = CreateAbTestForm(request.POST)

        if form.is_valid():
            ab_test = form.save(page, page.get_latest_revision(), request.user)

            if "start" in request.POST:
                page_perms = page.permissions_for_user(request.user)
                if page_perms.can_publish():
                    ab_test.start()

                    messages.success(
                        request, _("The A/B test has been created and started.")
                    )
                else:
                    messages.error(
                        request,
                        _(
                            "The A/B test has been created but you do not have permission to publish so it couldn't be started."
                        ),
                    )

            else:
                messages.success(request, _("The A/B test has been created."))

            return redirect("wagtailadmin_pages:edit", page.id)
    else:
        form = CreateAbTestForm()

    event_types = get_event_types().items()

    panel = panels.ObjectList(form.panels).bind_to_model(AbTest)
    bound_panel = panel.get_bound_panel(
        instance=form.instance, form=form, request=request
    )

    """
    Template: wagtail_ab_testing/add_form.html is rendered here

    Passing this to the template so we can test for it where an include is used because we can't use
    "wagtailadmin/pages/_editor_css.html" as it's not available in Wagtail 4+
    """

    return render(
        request,
        "wagtail_ab_testing/add_form.html",
        {
            "page": page,
            "form": form,
            "panel": bound_panel,
            "goal_selector_props": json.dumps(
                {
                    "goalTypesByPageType": {
                        f"{page_type._meta.app_label}.{page_type._meta.model_name}": [
                            {
                                "slug": slug,
                                "name": event_type.name,
                            }
                            for slug, event_type in event_types
                            if event_type.requires_page
                            and event_type.can_be_triggered_on_page_type(page_type)
                        ]
                        for page_type in PAGE_MODEL_CLASSES
                    },
                    "globalGoalTypes": [
                        {
                            "slug": slug,
                            "name": event_type.name,
                        }
                        for slug, event_type in event_types
                        if not event_type.requires_page
                    ],
                },
                cls=DjangoJSONEncoder,
            ),
        },
    )


class StartAbTestMenuItem(ActionMenuItem):
    name = "action-start-ab-test"
    label = _("Start A/B test")

    def is_shown(self, request, context):
        page = context["ab_test"].page
        if not page.permissions_for_user(request.user).can_publish():
            return False

        return context["ab_test"].status == AbTest.STATUS_DRAFT


class RestartAbTestMenuItem(ActionMenuItem):
    name = "action-restart-ab-test"
    label = _("Restart A/B test")

    def is_shown(self, request, context):
        page = context["ab_test"].page
        if not page.permissions_for_user(request.user).can_publish():
            return False

        return context["ab_test"].status == AbTest.STATUS_PAUSED


class EndAbTestMenuItem(ActionMenuItem):
    name = "action-end-ab-test"
    label = _("End A/B test")

    def is_shown(self, request, context):
        page = context["ab_test"].page
        if not page.permissions_for_user(request.user).can_publish():
            return False

        return context["ab_test"].status in [
            AbTest.STATUS_DRAFT,
            AbTest.STATUS_RUNNING,
            AbTest.STATUS_PAUSED,
        ]


class PauseAbTestMenuItem(ActionMenuItem):
    name = "action-pause-ab-test"
    label = _("Pause A/B test")

    def is_shown(self, request, context):
        page = context["ab_test"].page
        if not page.permissions_for_user(request.user).can_publish():
            return False

        return context["ab_test"].status == AbTest.STATUS_RUNNING


class AbTestActionMenu:
    template = "wagtailadmin/pages/action_menu/menu.html"

    def __init__(self, request, **kwargs):
        self.request = request
        self.context = kwargs
        # The ActionMenuItem request object is available in the context dictionary as context['request'].
        # https://docs.wagtail.io/en/stable/releases/2.15.html#admin-homepage-panels-summary-items-and-action-menu-items-now-use-components
        self.context["request"] = request
        self.menu_items = [
            StartAbTestMenuItem(order=0),
            RestartAbTestMenuItem(order=1),
            EndAbTestMenuItem(order=2),
            PauseAbTestMenuItem(order=3),
        ]

        self.menu_items = [
            menu_item
            for menu_item in self.menu_items
            if menu_item.is_shown(self.request, self.context)
        ]

        try:
            self.default_item = self.menu_items.pop(0)
        except IndexError:
            self.default_item = None

    def render_html(self):
        return render_to_string(
            self.template,
            {
                "default_menu_item": self.default_item.render_html(self.context),
                "show_menu": bool(self.menu_items),
                "rendered_menu_items": [
                    menu_item.render_html(self.context)
                    for menu_item in self.menu_items
                ],
            },
            request=self.request,
        )

    @cached_property
    def media(self):
        media = forms.Media()
        for item in self.menu_items:
            media += item.media
        return media


def get_progress_and_results_common_context(request, page, ab_test):
    # Fetch stats from database
    stats = ab_test.hourly_logs.aggregate(
        control_participants=Sum(
            "participants", filter=Q(version=AbTest.VERSION_CONTROL)
        ),
        control_conversions=Sum(
            "conversions", filter=Q(version=AbTest.VERSION_CONTROL)
        ),
        variant_participants=Sum(
            "participants", filter=Q(version=AbTest.VERSION_VARIANT)
        ),
        variant_conversions=Sum(
            "conversions", filter=Q(version=AbTest.VERSION_VARIANT)
        ),
    )
    control_participants = stats["control_participants"] or 0
    control_conversions = stats["control_conversions"] or 0
    variant_participants = stats["variant_participants"] or 0
    variant_conversions = stats["variant_conversions"] or 0

    current_sample_size = control_participants + variant_participants

    estimated_completion_date = None
    if ab_test.status == AbTest.STATUS_RUNNING and current_sample_size:
        running_duration_days = ab_test.total_running_duration().days

        if running_duration_days > 0:
            participants_per_day = (
                current_sample_size / ab_test.total_running_duration().days
            )
            estimated_days_remaining = (
                ab_test.sample_size - current_sample_size
            ) / participants_per_day
            estimated_completion_date = timezone.now().date() + datetime.timedelta(
                days=estimated_days_remaining
            )

    # Generate time series data for the chart
    time_series = []
    control = 0
    variant = 0
    date = None
    for log in ab_test.hourly_logs.order_by("date", "hour"):
        # Accumulate the conversions
        if log.version == AbTest.VERSION_CONTROL:
            control += log.conversions
        else:
            variant += log.conversions

        while date is None or date < log.date:
            if date is None:
                # First record
                date = log.date
            else:
                # Move time forward to match log record
                date += datetime.timedelta(days=1)

            # Generate a log for this time
            time_series.append(
                {
                    "date": date,
                    "control": control,
                    "variant": variant,
                }
            )

    # Format stats for display
    control_conversions_percent = formats.localize(
        round(control_conversions / control_participants * 100, 1)
    ) if control_participants else 0
    variant_conversions_percent = formats.localize(
        round(variant_conversions / variant_participants * 100, 1)
    ) if variant_conversions else 0

    return {
        "page": page,
        "ab_test": ab_test,
        "current_sample_size": current_sample_size,
        "current_sample_size_percent": int(
            current_sample_size / ab_test.sample_size * 100
        ),
        "control_conversions": control_conversions,
        "control_participants": control_participants,
        "control_conversions_percent": control_conversions_percent,
        "variant_conversions": variant_conversions,
        "variant_participants": variant_participants,
        "variant_conversions_percent": variant_conversions_percent,
        "control_is_winner": ab_test.winning_version == AbTest.VERSION_CONTROL,
        "variant_is_winner": ab_test.winning_version == AbTest.VERSION_VARIANT,
        "unclear_winner": ab_test.status
        in [AbTest.STATUS_FINISHED, ab_test.STATUS_COMPLETED]
        and ab_test.winning_version is None,
        "estimated_completion_date": estimated_completion_date,
        "chart_data": json.dumps(
            {
                "x": "x",
                "columns": [
                    ["x"]
                    + [data_point["date"].isoformat() for data_point in time_series],
                    [_("Control")]
                    + [data_point["control"] for data_point in time_series],
                    [_("Variant")]
                    + [data_point["variant"] for data_point in time_series],
                ],
                "type": "spline",
            }
        ),
    }


def progress(request, page, ab_test):
    if request.method == "POST":
        page_perms = page.permissions_for_user(request.user)

        if (
            "action-start-ab-test" in request.POST
            or "action-restart-ab-test" in request.POST
        ):
            if page_perms.can_publish():
                if ab_test.status in [AbTest.STATUS_DRAFT, AbTest.STATUS_PAUSED]:
                    ab_test.start()

                    messages.success(request, _("The A/B test has been started."))
                else:
                    messages.error(
                        request,
                        _(
                            "The A/B test must be in draft or paused in order to be started."
                        ),
                    )
            else:
                messages.error(
                    request,
                    _(
                        "You must have permission to publish in order to start an A/B test."
                    ),
                )

        elif "action-end-ab-test" in request.POST:
            if page_perms.can_publish():
                if ab_test.status in [
                    AbTest.STATUS_DRAFT,
                    AbTest.STATUS_RUNNING,
                    AbTest.STATUS_PAUSED,
                ]:
                    ab_test.cancel()

                    messages.success(request, _("The A/B test has been ended."))
                elif ab_test.status == AbTest.STATUS_FINISHED:
                    ab_test.complete(
                        AbTest.COMPLETION_ACTION_DO_NOTHING, user=request.user
                    )
                else:
                    messages.error(request, _("The A/B test has already ended."))
            else:
                messages.error(
                    request,
                    _(
                        "You must have permission to publish in order to end an A/B test."
                    ),
                )

        elif "action-pause-ab-test" in request.POST:
            if page_perms.can_publish():
                if ab_test.status == AbTest.STATUS_RUNNING:
                    ab_test.pause()

                    messages.success(request, _("The A/B test has been paused."))
                else:
                    messages.error(
                        request,
                        _("The A/B test cannot be paused because it is not running."),
                    )
            else:
                messages.error(
                    request,
                    _(
                        "You must have permission to publish in order to pause an A/B test."
                    ),
                )

        elif "action-select-control" in request.POST:
            if ab_test.status == AbTest.STATUS_FINISHED:
                ab_test.complete(AbTest.COMPLETION_ACTION_REVERT, user=request.user)

                messages.success(
                    request,
                    _("The page has been reverted back to the control version."),
                )

            else:
                messages.error(
                    request,
                    _("The A/B test cannot be paused because it is not running."),
                )

        elif "action-select-variant" in request.POST:
            if ab_test.status == AbTest.STATUS_FINISHED:
                # TODO Permission check?
                ab_test.complete(AbTest.COMPLETION_ACTION_PUBLISH, user=request.user)

                messages.success(request, _("The variant version has been published."))

            else:
                messages.error(
                    request,
                    _("The A/B test cannot be paused because it is not running."),
                )

        else:
            messages.error(request, _("Unknown action"))

        # Redirect back
        return redirect("wagtailadmin_pages:edit", page.id)

    context = get_progress_and_results_common_context(request, page, ab_test)
    context["action_menu"] = AbTestActionMenu(
        request, view="edit", page=page, ab_test=ab_test
    )
    return render(request, "wagtail_ab_testing/progress.html", context)


def results(request, page_id, ab_test_id):
    page = get_object_or_404(Page, id=page_id)
    if not page.permissions_for_user(request.user).can_edit():
        raise PermissionDenied

    ab_test = get_object_or_404(AbTest, page=page, id=ab_test_id)

    context = get_progress_and_results_common_context(request, page, ab_test)
    return render(request, "wagtail_ab_testing/results.html", context)


def compare_draft(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    latest_revision_as_page = page.get_latest_revision_as_object()

    comparison = (
        page.get_edit_handler()
        .get_bound_panel(page, request=request, form=None)
        .get_comparison()
    )

    comparison = [comp(page, latest_revision_as_page) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return render(
        request,
        "wagtail_ab_testing/compare.html",
        {
            "page": page,
            "latest_revision_as_page": latest_revision_as_page,
            "comparison": comparison,
        },
    )


class SearchPageTitleFilter(django_filters.CharFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return qs.filter(page__title__icontains=value)


class AbTestingReportFilterSet(WagtailFilterSet):
    page = SearchPageTitleFilter()
    first_started_at = django_filters.DateFromToRangeFilter(
        label=gettext_lazy("Started at"), widget=DateRangePickerWidget
    )

    class Meta:
        model = AbTest
        fields = ["status", "page", "first_started_at"]


class AbTestingReportView(ReportView):
    template_name = "wagtail_ab_testing/report.html"
    title = gettext_lazy("A/B testing")
    header_icon = ""

    filterset_class = AbTestingReportFilterSet

    def get_queryset(self):
        return AbTest.objects.all().order_by(
            F("first_started_at").desc(nulls_first=True)
        )


@csrf_exempt
@api_view(["POST"])
@authentication_classes([])
@permission_classes([])
def register_participant(request):
    test_id = request.data.get("test_id", None)
    if test_id is None:
        return Response("test_id not provided", status=status.HTTP_400_BAD_REQUEST)

    try:
        test_id = int(test_id)
        if test_id < 1:
            raise ValueError
    except ValueError:
        return Response(
            "test_id must be a positive integer", status=status.HTTP_400_BAD_REQUEST
        )

    test = get_object_or_404(AbTest, id=test_id)

    version = request.data.get("version", None)
    if version is None:
        return Response("version not provided", status=status.HTTP_400_BAD_REQUEST)

    if version not in [AbTest.VERSION_CONTROL, AbTest.VERSION_VARIANT]:
        return Response(
            "version must be either '{}' or '{}'".format(
                AbTest.VERSION_CONTROL, AbTest.VERSION_VARIANT
            ),
            status=status.HTTP_400_BAD_REQUEST,
        )

    # Add participant
    test.add_participant(version=version)

    return Response()


@csrf_exempt
@api_view(["POST"])
@authentication_classes([])
@permission_classes([])
def goal_reached(request):
    test_id = request.data.get("test_id", None)
    if test_id is None:
        return Response("test_id not provided", status=status.HTTP_400_BAD_REQUEST)

    test = get_object_or_404(AbTest, id=test_id)

    version = request.data.get("version", None)
    if version is None:
        return Response("version not provided", status=status.HTTP_400_BAD_REQUEST)

    if version not in [AbTest.VERSION_CONTROL, AbTest.VERSION_VARIANT]:
        return Response(
            "version must be either '{}' or '{}'".format(
                AbTest.VERSION_CONTROL, AbTest.VERSION_VARIANT
            ),
            status=status.HTTP_400_BAD_REQUEST,
        )

    # Log conversion
    test.log_conversion(version)

    return Response()



# Generated by Django 3.1.3 on 2020-12-14 17:10

from django.db import migrations


def grant_moderators_add_abtest_permission(apps, schema_editor):
    ContentType = apps.get_model('contenttypes.ContentType')
    Permission = apps.get_model('auth.Permission')
    Group = apps.get_model('auth.Group')

    abtest_content_type, created = ContentType.objects.get_or_create(app_label='wagtail_ab_testing', model='abtest')
    add_abtest_permission, created = Permission.objects.get_or_create(content_type=abtest_content_type, codename='add_abtest')

    moderators_group = Group.objects.filter(name='Moderators').first()
    if moderators_group:
        moderators_group.permissions.add(add_abtest_permission)


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0006_sample_size_min_value'),
    ]

    operations = [
        migrations.RunPython(grant_moderators_add_abtest_permission, migrations.RunPython.noop),
    ]



# Generated by Django 3.1.3 on 2020-11-24 16:05

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('wagtail_ab_testing', '0004_started_at_and_duration'),
    ]

    operations = [
        migrations.AddField(
            model_name='abtest',
            name='created_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='abtest',
            name='hypothesis',
            field=models.TextField(blank=True),
        ),
    ]



from django import template
from django.urls import reverse

from wagtail_ab_testing.models import AbTest
from wagtail_ab_testing.utils import request_is_trackable

register = template.Library()


@register.inclusion_tag('wagtail_ab_testing/script.html', takes_context=True)
def wagtail_ab_testing_script(context):
    request = context['request']
    serving_variant = getattr(request, 'wagtail_ab_testing_serving_variant', False)

    track = request_is_trackable(request)
    if not track:
        return {
            "track": False,
            "tracking_parameters": None,
        }

    register_participant_url = reverse('wagtail_ab_testing:register_participant')
    goal_reached_url = reverse('wagtail_ab_testing:goal_reached')

    tracking_parameters = {
        "urls": {
            "registerParticipant": register_participant_url,
            "goalReached": goal_reached_url,
        },
    }

    page = context.get('page', None)
    page_id = page.id if page else None
    if page_id:
        tracking_parameters["pageId"] = page_id

    version = AbTest.VERSION_VARIANT if serving_variant else AbTest.VERSION_CONTROL
    test = getattr(request, 'wagtail_ab_testing_test', None)

    if (test and version):
        tracking_parameters["testId"] = test.id
        tracking_parameters["version"] = version
        tracking_parameters["goalEvent"] = test.goal_event
        tracking_parameters["goalPageId"] = test.goal_page.id if test.goal_page else None

    return {
        'track': track,
        'tracking_parameters': tracking_parameters,
    }



# Generated by Django 4.1.6 on 2023-03-02 13:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0078_referenceindex'),
        ('wagtail_ab_testing', '0011_rename_treatment_to_variant_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='abtest',
            name='variant_revision',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.revision'),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-05 18:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0002_abtesthourlylog'),
    ]

    operations = [
        migrations.AddField(
            model_name='abtest',
            name='winning_variant',
            field=models.CharField(choices=[('control', 'Control'), ('treatment', 'Treatment')], max_length=9, null=True),
        ),
    ]



import os

if os.name == 'nt':
    # Windows has a different strftime format for dates without leading 0
    # https://stackoverflow.com/questions/904928/python-strftime-date-without-leading-0
    DATE_FORMAT = '%#d %B %Y'
else:
    DATE_FORMAT = '%-d %B %Y'



import random

from datetime import datetime, timedelta, timezone as tz

import scipy.stats
import numpy as np
from django.conf import settings
from django.core.validators import MinValueValidator
from django.db import connection, models, transaction
from django.db.models import Q, Sum
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext as _, gettext_lazy as __

from wagtail.signals import page_unpublished

from .events import get_event_types


class AbTestManager(models.Manager):
    def get_current_for_page(self, page):
        return self.get_queryset().filter(page=page).exclude(status__in=[AbTest.STATUS_CANCELLED, AbTest.STATUS_COMPLETED]).first()


class AbTest(models.Model):
    """
    Represents an A/B test that has been set up by the user.

    The live page content is used as the control, the revision pointed to in
    the `.variant_revision` field contains the changes that are being tested.
    """

    STATUS_DRAFT = 'draft'
    STATUS_RUNNING = 'running'
    STATUS_PAUSED = 'paused'
    STATUS_CANCELLED = 'cancelled'
    # These two sound similar, but there's a difference:
    # 'Finished' means that we've reached the sample size and testing has stopped
    # but the user still needs to decide whether to publish the variant version
    # or revert back to the control.
    # Once they've decided and that action has taken place, the test status is
    # updated to 'Completed'.
    STATUS_FINISHED = 'finished'
    STATUS_COMPLETED = 'completed'

    STATUS_CHOICES = [
        (STATUS_DRAFT, __('Draft')),
        (STATUS_RUNNING, __('Running')),
        (STATUS_PAUSED, __('Paused')),
        (STATUS_CANCELLED, __('Cancelled')),
        (STATUS_FINISHED, __('Finished')),
        (STATUS_COMPLETED, __('Completed')),
    ]

    VERSION_CONTROL = 'control'
    VERSION_VARIANT = 'variant'

    VERSION_CHOICES = [
        (VERSION_CONTROL, __('Control')),
        (VERSION_VARIANT, __('Variant')),
    ]

    COMPLETION_ACTION_DO_NOTHING = 'do-nothing'
    COMPLETION_ACTION_REVERT = 'revert'
    COMPLETION_ACTION_PUBLISH = 'publish'

    COMPLETION_ACTION_CHOICES = [
        (COMPLETION_ACTION_DO_NOTHING, "Do nothing"),
        (COMPLETION_ACTION_REVERT, "Revert"),
        (COMPLETION_ACTION_PUBLISH, "Publish"),
    ]

    page = models.ForeignKey('wagtailcore.Page', on_delete=models.CASCADE, related_name='ab_tests')
    name = models.CharField(max_length=255)
    hypothesis = models.TextField(blank=True)
    variant_revision = models.ForeignKey('wagtailcore.Revision', on_delete=models.PROTECT, related_name='+')
    goal_event = models.CharField(max_length=255)
    goal_page = models.ForeignKey('wagtailcore.Page', null=True, blank=True, on_delete=models.SET_NULL, related_name='+')
    sample_size = models.PositiveIntegerField(validators=[MinValueValidator(1)])
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, blank=True, related_name='+')
    status = models.CharField(max_length=20, choices=STATUS_CHOICES, default=STATUS_DRAFT)
    winning_version = models.CharField(max_length=9, null=True, choices=VERSION_CHOICES)
    first_started_at = models.DateTimeField(null=True)

    # Because an admin can pause/resume tests, we need to make sure we record the amount of time it has been running
    previous_run_duration = models.DurationField(default=timedelta(0))
    current_run_started_at = models.DateTimeField(null=True)

    objects = AbTestManager()

    def get_goal_event_display(self):
        """
        Returns the display name of the goal event.
        """
        for event_type_slug, event_type in get_event_types().items():
            if event_type_slug == self.goal_event:
                return event_type.name

        return self.goal_event

    def start(self):
        """
        Starts/unpauses the test.
        """
        if self.status in [self.STATUS_DRAFT, self.STATUS_PAUSED]:
            self.current_run_started_at = timezone.now()

            if self.status == self.STATUS_DRAFT:
                self.first_started_at = self.current_run_started_at

            self.status = self.STATUS_RUNNING

            self.save(update_fields=['status', 'current_run_started_at', 'first_started_at'])

    def pause(self):
        """
        Pauses the test.
        """
        if self.status == self.STATUS_RUNNING:
            self.status = self.STATUS_PAUSED

            if self.current_run_started_at is not None:
                self.previous_run_duration += timezone.now() - self.current_run_started_at
                self.current_run_started_at = None

            self.save(update_fields=['status', 'previous_run_duration', 'current_run_started_at'])

    def get_results_url(self):
        """
        Returns the URL to the page wherethe user can see the results.

        While the test is running, this is the URL of the edit view.
        Afterwards, we need to send them to a separate view as the
        page editor returns to normal.
        """
        if self.status in [AbTest.STATUS_COMPLETED, AbTest.STATUS_CANCELLED]:
            return reverse('wagtail_ab_testing_admin:results', args=[self.page_id, self.id])

        else:
            return reverse('wagtailadmin_pages:edit', args=[self.page_id])

    def total_running_duration(self):
        """
        Returns the total duration that this test has been running.
        """
        duration = self.previous_run_duration

        if self.status == self.STATUS_RUNNING:
            duration += timezone.now() - self.current_run_started_at

        return duration

    def cancel(self):
        """
        Cancels the test.
        """
        self.status = self.STATUS_CANCELLED

        self.save(update_fields=['status'])

    def finish(self):
        """
        Finishes the testing.

        Note that this doesn't 'complete' the test: a finished test means
        that testing is no longer happening. The test is not complete until
        the user decides on the outcome of the test (keep the control or
        publish the variant). This decision is set using the .complete()
        method.
        """
        self.status = self.STATUS_FINISHED
        self.winning_version = self.check_for_winner()

        self.save(update_fields=['status', 'winning_version'])

    @transaction.atomic
    def complete(self, action, user=None):
        """
        Completes the test and carries out the specificed action.

        Actions can be:
         - AbTest.COMPLETION_ACTION_DO_NOTHING - This just completes
           the test but does nothing to the page. The control will
           remain the published version and the variant will be
           in draft.
         - AbTest.COMPLETION_ACTION_REVERT - This completes the test
           and also creates a new revision to revert the content back
           to what it was in the control while the test was taking
           place.
         - AbTest.COMPLETION_ACTION_PUBLISH - This completes the test
           and also publishes the variant revision.
        """
        self.status = self.STATUS_COMPLETED
        self.save(update_fields=['status'])

        if action == AbTest.COMPLETION_ACTION_DO_NOTHING:
            pass

        elif action == AbTest.COMPLETION_ACTION_REVERT:
            # Create a new revision with the content of the live page and publish it
            self.page.specific.save_revision(user=user, previous_revision=self.page.live_revision).publish(user=user)

        elif action == AbTest.COMPLETION_ACTION_PUBLISH:
            self.variant_revision.publish(user=user)

    def get_participation_numbers(self):
        """
        Returns a 2-tuple containing the number of participants who were given the control or variant version of the page respectively.
        """
        stats = self.hourly_logs.aggregate(
            control_participants=Sum('participants', filter=Q(version=self.VERSION_CONTROL)),
            variant_participants=Sum('participants', filter=Q(version=self.VERSION_VARIANT)),
        )
        control_participants = stats['control_participants'] or 0
        variant_participants = stats['variant_participants'] or 0

        return control_participants, variant_participants

    def get_new_participant_version(self, participation_numbers=None):
        """
        Returns the version of the page to display to a new participant.
        This balances the number of participants between the two variants.
        """
        if participation_numbers is None:
            participation_numbers = self.get_participation_numbers()

        control_participants, variant_participants = participation_numbers

        if variant_participants > control_participants:
            return self.VERSION_CONTROL

        elif variant_participants < control_participants:
            return self.VERSION_VARIANT

        else:
            return random.choice([
                self.VERSION_CONTROL,
                self.VERSION_VARIANT,
            ])

    def add_participant(self, version=None):
        """
        Inserts a new participant into the log. Returns the version that they should be shown.
        """
        # Get current numbers of participants for each version
        control_participants, variant_participants = self.get_participation_numbers()

        # Create an equal number of participants for each version
        if version is None:
            # Note, pass participation numbers we already have to save a database query
            version = self.get_new_participant_version(participation_numbers=(control_participants, variant_participants))

        # Add new participant to statistics model
        AbTestHourlyLog._increment_stats(self, version, 1, 0)

        # If we have now reached the required sample size, end the test
        # Note: we don't care too much that the last few participants won't
        # get a chance to turn into conversions. It's unlikely to make a
        # significant difference to the results.
        # Note: Adding 1 to account for the new participant
        if control_participants + variant_participants + 1 >= self.sample_size:
            self.finish()

        return version

    def log_conversion(self, version, *, time=None):
        """
        Logs when a participant completed the goal.

        Note: It's up to the caller to make sure that this doesn't get called more than once
        per participant.
        """
        AbTestHourlyLog._increment_stats(self, version, 0, 1, time=time)

    def check_for_winner(self):
        """
        Performs a Chi-Squared test to check if there is a clear winner.

        Returns VERSION_CONTROL or VERSION_VARIANT if there is one. Otherwise, it returns None.

        For more information on what the Chi-Squared test does, see:
        https://www.evanmiller.org/ab-testing/chi-squared.html
        https://towardsdatascience.com/a-b-testing-with-chi-squared-test-to-maximize-conversions-and-ctrs-6599271a2c31
        """
        # Fetch stats from database
        stats = self.hourly_logs.aggregate(
            control_participants=Sum('participants', filter=Q(version=self.VERSION_CONTROL)),
            control_conversions=Sum('conversions', filter=Q(version=self.VERSION_CONTROL)),
            variant_participants=Sum('participants', filter=Q(version=self.VERSION_VARIANT)),
            variant_conversions=Sum('conversions', filter=Q(version=self.VERSION_VARIANT)),
        )
        control_participants = stats['control_participants'] or 0
        control_conversions = stats['control_conversions'] or 0
        variant_participants = stats['variant_participants'] or 0
        variant_conversions = stats['variant_conversions'] or 0

        if not control_conversions and not variant_conversions:
            return

        if control_conversions > control_participants or variant_conversions > variant_participants:
            # Something's up. I'm sure it's already clear in the UI what's going on, so let's not crash
            return

        # Create a numpy array with values to pass in to Chi-Squared test
        control_failures = control_participants - control_conversions
        variant_failures = variant_participants - variant_conversions

        if control_failures == 0 and variant_failures == 0:
            # Prevent this error: "The internally computed table of expected frequencies has a zero element at (0, 1)."
            return

        T = np.array([[control_conversions, control_failures], [variant_conversions, variant_failures]])

        # Perform Chi-Squared test
        p = scipy.stats.chi2_contingency(T, correction=False)[1]

        # Check if there is a clear winner
        required_confidence_level = 0.95  # 95%
        if 1 - p > required_confidence_level:
            # There is a clear winner!
            # Return the one with the highest success rate
            if (control_conversions / control_participants) > (variant_conversions / variant_participants):
                return self.VERSION_CONTROL
            else:
                return self.VERSION_VARIANT

    def get_status_description(self):
        """
        Returns a string that describes the status in more detail.
        """
        status = self.get_status_display()

        if self.status == AbTest.STATUS_RUNNING:
            participants = self.hourly_logs.aggregate(participants=Sum('participants'))['participants'] or 0
            completeness_percentange = int((participants * 100) / self.sample_size)
            return status + f" ({completeness_percentange}%)"

        elif self.status in [AbTest.STATUS_FINISHED, AbTest.STATUS_COMPLETED]:
            if self.winning_version == AbTest.VERSION_CONTROL:
                return status + " (" + _("Control won") + ")"

            elif self.winning_version == AbTest.VERSION_VARIANT:
                return status + " (" + _("Variant won") + ")"

            else:
                return status + " (" + _("No clear winner") + ")"

        else:
            return status


class AbTestHourlyLog(models.Model):
    ab_test = models.ForeignKey(AbTest, on_delete=models.CASCADE, related_name='hourly_logs')
    version = models.CharField(max_length=9, choices=AbTest.VERSION_CHOICES)
    date = models.DateField()
    # UTC hour. Values range from 0 to 23
    hour = models.PositiveSmallIntegerField()

    # New participants added in this hour
    participants = models.PositiveIntegerField(default=0)

    # New or existing participants that converted in this hour
    conversions = models.PositiveIntegerField(default=0)

    @classmethod
    def _increment_stats(cls, ab_test, version, participants, conversions, *, time=None):
        """
        Increments the participants/conversions statistics for the given ab_test/version.

        This will create a new AbTestHourlyLog record if one doesn't exist for the current hour.
        """
        time = time.astimezone(tz.utc) if time else datetime.utcnow()
        date = time.date()
        hour = time.hour

        if connection.vendor == 'postgresql':
            # Use fast, atomic UPSERT query on PostgreSQL
            with connection.cursor() as cursor:
                table_name = connection.ops.quote_name(cls._meta.db_table)
                query = """
                    INSERT INTO %s (ab_test_id, version, date, hour, participants, conversions)
                    VALUES (%%s, %%s, %%s, %%s, %%s, %%s)
                    ON CONFLICT (ab_test_id, version, date, hour)
                        DO UPDATE SET participants = %s.participants + %%s, conversions = %s.conversions + %%s;
                """ % (table_name, table_name, table_name)

                cursor.execute(query, [
                    ab_test.id,
                    version,
                    date,
                    hour,
                    participants,
                    conversions,
                    participants,
                    conversions
                ])
        else:
            # Fall back to running two queries (with small potential for race conditions if things run slowly)
            hourly_log, created = cls.objects.get_or_create(
                ab_test=ab_test,
                version=version,
                date=date,
                hour=hour,
                defaults={
                    'participants': participants,
                    'conversions': conversions,
                }
            )

            if not created:
                hourly_log.participants += participants
                hourly_log.conversions += conversions
                hourly_log.save(update_fields=['participants', 'conversions'])

    class Meta:
        ordering = ['ab_test', 'version', 'date', 'hour']
        unique_together = [
            ('ab_test', 'version', 'date', 'hour'),
        ]


@receiver(page_unpublished)
def cancel_on_page_unpublish(instance, **kwargs):
    for ab_test in AbTest.objects.filter(page=instance, status__in=[AbTest.STATUS_DRAFT, AbTest.STATUS_RUNNING, AbTest.STATUS_PAUSED]):
        ab_test.cancel()

    for ab_test in AbTest.objects.filter(page=instance, status=AbTest.STATUS_FINISHED):
        ab_test.complete(AbTest.COMPLETION_ACTION_DO_NOTHING)



# Generated by Django 3.1.3 on 2020-11-05 18:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AbTestHourlyLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('variant', models.CharField(choices=[('control', 'Control'), ('treatment', 'Treatment')], max_length=9)),
                ('date', models.DateField()),
                ('hour', models.PositiveSmallIntegerField()),
                ('participants', models.PositiveIntegerField(default=0)),
                ('conversions', models.PositiveIntegerField(default=0)),
                ('ab_test', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='hourly_logs', to='wagtail_ab_testing.abtest')),
            ],
            options={
                'ordering': ['ab_test', 'variant', 'date', 'hour'],
                'unique_together': {('ab_test', 'variant', 'date', 'hour')},
            },
        ),
    ]



from functools import lru_cache

from user_agents import parse


@lru_cache(maxsize=1000)
def is_bot(ua_string):
    return parse(ua_string).is_bot


def request_is_trackable(request):
    """
    Returns True if we can track the specified request.
    """
    # Don't track anyone with HTTP Do-Not-Track enabled
    if 'HTTP_DNT' in request.META and request.META['HTTP_DNT'] == '1':
        return False

    # Don't track bots
    ua_string = request.META.get('HTTP_USER_AGENT', '')
    if not isinstance(ua_string, str):
        ua_string = ua_string.decode('utf-8', 'ignore')
    if is_bot(ua_string):
        return False

    return True



# Generated by Django 3.1.3 on 2020-11-03 15:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0052_pagelogentry'),
    ]

    operations = [
        migrations.CreateModel(
            name='AbTest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('goal_event', models.CharField(max_length=255)),
                ('sample_size', models.PositiveIntegerField()),
                ('status', models.CharField(choices=[('draft', 'Draft'), ('running', 'Running'), ('paused', 'Paused'), ('cancelled', 'Cancelled'), ('completed', 'Completed')], default='draft', max_length=20)),
                ('goal_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.page')),
                ('page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ab_tests', to='wagtailcore.page')),
                ('treatment_revision', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.pagerevision')),
            ],
        ),
    ]



from django.urls import reverse
from rest_framework import fields, routers, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from wagtail.models import Page, Site

from .models import AbTest


class SiteSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ['id', 'hostname']
        model = Site


class PageSerializer(serializers.ModelSerializer):
    path = fields.SerializerMethodField('get_path')

    def get_path(self, page):
        return page.get_url_parts()[2]

    class Meta:
        fields = ['id', 'path']
        model = Page


class AbTestGoalSerializer(serializers.ModelSerializer):
    page = PageSerializer(source='goal_page')
    event = fields.ReadOnlyField(source='goal_event')

    class Meta:
        fields = ['page', 'event']
        model = AbTest


class AbTestSerializer(serializers.ModelSerializer):
    site = SiteSerializer(source='page.get_site')
    page = PageSerializer()
    goal = AbTestGoalSerializer(source='*')
    variant_html_url = fields.SerializerMethodField()
    add_participant_url = fields.SerializerMethodField()
    log_conversion_url = fields.SerializerMethodField()

    def get_variant_html_url(self, test):
        return reverse('wagtail_ab_testing_api:abtest-serve-variant', args=[test.id])

    def get_add_participant_url(self, test):
        return reverse('wagtail_ab_testing_api:abtest-add-participant', args=[test.id])

    def get_log_conversion_url(self, test):
        return reverse('wagtail_ab_testing_api:abtest-log-conversion', args=[test.id])

    class Meta:
        fields = ['id', 'site', 'page', 'goal', 'variant_html_url', 'add_participant_url', 'log_conversion_url']
        model = AbTest


class AbTestViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = AbTestSerializer
    queryset = AbTest.objects.filter(status=AbTest.STATUS_RUNNING)

    @action(detail=True, methods=['get'])
    def serve_variant(self, request, pk=None):
        test = self.get_object()
        request.wagtail_ab_testing_test = test
        request.wagtail_ab_testing_serving_variant = True
        return test.variant_revision.as_object().serve(request)

    @action(detail=True, methods=['post'])
    def add_participant(self, request, pk=None):
        test = self.get_object()
        variant = test.add_participant()
        return Response({
            'version': variant,
            'test_finished': test.status != AbTest.STATUS_RUNNING,
        }, status=status.HTTP_201_CREATED)

    @action(detail=True, methods=['post'])
    def log_conversion(self, request, pk=None):
        test = self.get_object()

        if request.data['version'] not in dict(AbTest.VERSION_CHOICES).keys():
            return Response({}, status=status.HTTP_400_BAD_REQUEST)

        test.log_conversion(request.data['version'])
        return Response({}, status=status.HTTP_201_CREATED)


router = routers.SimpleRouter()
router.register(r'tests', AbTestViewSet)

app_name = 'wagtail_ab_testing_api'
urlpatterns = router.urls



from django.utils.translation import gettext_lazy as __

from wagtail import hooks


class BaseEvent:
    """
    A base class for events that are linked to Wagtail pages.
    """
    name = None

    # When False, the user won't be asked to select a goal page
    requires_page = True

    # A list of page model classes that this event can be triggered on.
    # This may be overridden by the .get_page_types() method
    # Leave this as None to allow any page
    page_types = None

    def get_page_types(self):
        """
        Returns a list of page model classes that this event can be triggered on.
        """
        return self.page_types

    def can_be_triggered_on_page_type(self, page_type):
        """
        Returns True if this event can be triggered on an instance of the given page model.
        """
        page_types = self.get_page_types()

        if page_types is None:
            return True

        return page_type in page_types


class VisitPageEvent(BaseEvent):
    """
    Triggered when a user visits a page.
    """
    name = __("Visit page")


# TODO: Find a way to hook this into Wagtail
# from wagtail.contrib.forms.utils import get_form_types
# class SubmitFormPageEvent(BaseEvent):
#     """
#     Triggered when a user submits a form page.
#     """
#     name = __("Submit form page")

#     def get_page_types(self):
#         return [
#             ct.model_class()
#             for ct in get_form_types()
#         ]


BUILTIN_EVENT_TYPES = {
    'visit-page': VisitPageEvent(),
    # 'submit-form': SubmitFormPageEvent(),
}


def get_event_types():
    event_types = {}
    event_types.update(BUILTIN_EVENT_TYPES)

    for fn in hooks.get_hooks('register_ab_testing_event_types'):
        event_types.update(fn())

    return event_types



# Generated by Django 3.1.3 on 2020-11-19 17:36

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0003_abtest_winning_variant'),
    ]

    operations = [
        migrations.AddField(
            model_name='abtest',
            name='current_run_started_at',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='abtest',
            name='first_started_at',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='abtest',
            name='previous_run_duration',
            field=models.DurationField(default=datetime.timedelta(0)),
        ),
    ]



# Generated by Django 3.1.3 on 2020-12-16 13:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0008_finished_status'),
    ]

    operations = [
        migrations.RenameField(
            model_name='abtest',
            old_name='winning_variant',
            new_name='winning_version',
        ),
        migrations.AlterModelOptions(
            name='abtesthourlylog',
            options={'ordering': ['ab_test', 'version', 'date', 'hour']},
        ),
        migrations.RenameField(
            model_name='abtesthourlylog',
            old_name='variant',
            new_name='version',
        ),
        migrations.AlterUniqueTogether(
            name='abtesthourlylog',
            unique_together={('ab_test', 'version', 'date', 'hour')},
        ),
    ]



# Generated by Django 3.1.3 on 2020-12-16 13:40

from django.db import migrations, models


def rename_treatment_to_variant_forwards(apps, schema_editor):
    AbTest = apps.get_model('wagtail_ab_testing.AbTest')
    AbTest.objects.filter(winning_version='treatment').update(winning_version='variant')

    AbTestHourlyLog = apps.get_model('wagtail_ab_testing.AbTestHourlyLog')
    AbTestHourlyLog.objects.filter(version='treatment').update(version='variant')


def rename_treatment_to_variant_backwards(apps, schema_editor):
    AbTest = apps.get_model('wagtail_ab_testing.AbTest')
    AbTest.objects.filter(winning_version='variant').update(winning_version='treatment')

    AbTestHourlyLog = apps.get_model('wagtail_ab_testing.AbTestHourlyLog')
    AbTestHourlyLog.objects.filter(version='variant').update(version='treatment')


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0010_rename_treatment_to_variant'),
    ]

    operations = [
        migrations.RunPython(rename_treatment_to_variant_forwards, rename_treatment_to_variant_backwards)
    ]



from django.apps import AppConfig


class WagtailAbTestingAppConfig(AppConfig):
    label = "wagtail_ab_testing"
    name = "wagtail_ab_testing"
    verbose_name = "A/B Testing"
    default_auto_field = "django.db.models.AutoField"



# Generated by Django 3.1.3 on 2020-12-09 11:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0007_grant_moderators_add_abtest_permission'),
    ]

    operations = [
        migrations.AlterField(
            model_name='abtest',
            name='status',
            field=models.CharField(choices=[('draft', 'Draft'), ('running', 'Running'), ('paused', 'Paused'), ('cancelled', 'Cancelled'), ('finished', 'Finished'), ('completed', 'Completed')], default='draft', max_length=20),
        ),
    ]



# Generated by Django 4.2.2 on 2023-06-18 11:26

# PURPOSE OF THIS MIGRATION
#
# This squashed migration was created for compatibility with Wagtail 4.0's migration from PageRevision 
# to a generic Revision model. The `0001_initial` migration in this project must continue to reference the old 
# PageRevision model and wagtailcore.0052_pagelogentry dependency to maintain a consistent migration history 
# for existing databases. On the flip side, new projects that run wagtail_ab_testing's migrations for the first time
# will encounter a PageRevision model that no longer exists which causes errors.
#
# The solution for new projects is to use this squashed migration file instead.
# This file contains no references to PageRevision and will migrate without errors.

import datetime
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.migrations.operations.special
import django.db.models.deletion


# Migrated from wagtail_ab_testing.migrations.0007_grant_moderators_add_abtest_permission
def grant_moderators_add_abtest_permission(apps, schema_editor):
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    Group = apps.get_model("auth.Group")

    abtest_content_type, created = ContentType.objects.get_or_create(
        app_label="wagtail_ab_testing", model="abtest"
    )
    add_abtest_permission, created = Permission.objects.get_or_create(
        content_type=abtest_content_type, codename="add_abtest"
    )

    moderators_group = Group.objects.filter(name="Moderators").first()
    if moderators_group:
        moderators_group.permissions.add(add_abtest_permission)


class Migration(migrations.Migration):
    replaces = [
        ("wagtail_ab_testing", "0001_initial"),
        ("wagtail_ab_testing", "0002_abtesthourlylog"),
        ("wagtail_ab_testing", "0003_abtest_winning_variant"),
        ("wagtail_ab_testing", "0004_started_at_and_duration"),
        ("wagtail_ab_testing", "0005_hypothesis_and_created_by"),
        ("wagtail_ab_testing", "0006_sample_size_min_value"),
        ("wagtail_ab_testing", "0007_grant_moderators_add_abtest_permission"),
        ("wagtail_ab_testing", "0008_finished_status"),
        ("wagtail_ab_testing", "0009_rename_variant_to_version"),
        ("wagtail_ab_testing", "0010_rename_treatment_to_variant"),
        ("wagtail_ab_testing", "0011_rename_treatment_to_variant_data"),
        ("wagtail_ab_testing", "0012_abtest_variant_revision"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("wagtailcore", "0078_referenceindex"),
    ]

    operations = [
        migrations.CreateModel(
            name="AbTest",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("goal_event", models.CharField(max_length=255)),
                (
                    "sample_size",
                    models.PositiveIntegerField(
                        validators=[django.core.validators.MinValueValidator(1)]
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("draft", "Draft"),
                            ("running", "Running"),
                            ("paused", "Paused"),
                            ("cancelled", "Cancelled"),
                            ("finished", "Finished"),
                            ("completed", "Completed"),
                        ],
                        default="draft",
                        max_length=20,
                    ),
                ),
                (
                    "goal_page",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="wagtailcore.page",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ab_tests",
                        to="wagtailcore.page",
                    ),
                ),
                (
                    "variant_revision",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="wagtailcore.revision",
                    ),
                ),
                (
                    "winning_version",
                    models.CharField(
                        choices=[("control", "Control"), ("variant", "Variant")],
                        max_length=9,
                        null=True,
                    ),
                ),
                ("current_run_started_at", models.DateTimeField(null=True)),
                ("first_started_at", models.DateTimeField(null=True)),
                (
                    "previous_run_duration",
                    models.DurationField(default=datetime.timedelta(0)),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("hypothesis", models.TextField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name="AbTestHourlyLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "version",
                    models.CharField(
                        choices=[("control", "Control"), ("variant", "Variant")],
                        max_length=9,
                    ),
                ),
                ("date", models.DateField()),
                ("hour", models.PositiveSmallIntegerField()),
                ("participants", models.PositiveIntegerField(default=0)),
                ("conversions", models.PositiveIntegerField(default=0)),
                (
                    "ab_test",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="hourly_logs",
                        to="wagtail_ab_testing.abtest",
                    ),
                ),
            ],
            options={"ordering": ["ab_test", "version", "date", "hour"]},
        ),
        migrations.AlterUniqueTogether(
            name="abtesthourlylog",
            unique_together={("ab_test", "version", "date", "hour")},
        ),
        migrations.RunPython(
            code=grant_moderators_add_abtest_permission,
            reverse_code=django.db.migrations.operations.special.RunPython.noop,
        ),
    ]



#!/usr/bin/env python

from setuptools import setup, find_packages

# Hack to prevent "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when setup.py exits
# (see http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="wagtail-ab-testing",
    version="0.10",
    description="A/B Testing for Wagtail",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Karl Hobley",
    author_email="karl@torchbox.com",
    url="",
    packages=find_packages(),
    include_package_data=True,
    license="BSD",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Wagtail",
        "Framework :: Wagtail :: 5",
        "Framework :: Wagtail :: 6",
    ],
    install_requires=[
        "Wagtail>=5.2",
        "user-agents>=2.2,<2.3",
        "numpy>=1.19.4,<2",
        "scipy>=1.5.4,<2",
    ],
    extras_require={
        "testing": ["dj-database-url==0.5.0", "freezegun==1.2.1"],
    },
    zip_safe=False,
)



from django.urls import path

from . import views

app_name = 'wagtail_ab_testing'
urlpatterns = [
    path('register-participant/', views.register_participant, name='register_participant'),
    path('goal-reached/', views.goal_reached, name='goal_reached'),
]



import json

from django.conf import settings
from django.contrib.auth.models import Permission
from django.core.exceptions import PermissionDenied
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import path, include, reverse
from django.utils.html import format_html, escapejs
from django.utils.translation import gettext as _, gettext_lazy as __
from django.views.i18n import JavaScriptCatalog


from wagtail import hooks
from wagtail.admin.action_menu import ActionMenuItem
from wagtail.admin.menu import MenuItem
from wagtail.admin.staticfiles import versioned_static

from . import views
from .compat import DATE_FORMAT
from .models import AbTest
from .utils import request_is_trackable


@hooks.register("register_admin_urls")
def register_admin_urls():
    urls = [
        path('jsi18n/', JavaScriptCatalog.as_view(packages=['wagtail_ab_testing']), name='javascript_catalog'),
        path('add/<int:page_id>/compare/', views.add_compare, name='add_ab_test_compare'),
        path('<int:page_id>/compare-draft/', views.compare_draft, name='compare_draft'),
        path('add/<int:page_id>/', views.add_form, name='add_ab_test_form'),
        path('report/', views.AbTestingReportView.as_view(), name='report'),
        path('results/<int:page_id>/<int:ab_test_id>/', views.results, name='results'),
    ]

    return [
        path(
            "abtests/",
            include(
                (urls, "wagtail_ab_testing_admin"),
                namespace="wagtail_ab_testing_admin",
            ),
        )
    ]


class CreateAbTestActionMenuItem(ActionMenuItem):
    name = 'create-ab-test'
    label = __("Save and create A/B Test")
    icon_name = 'people-arrows'

    def is_shown(self, context):
        if context['view'] != 'edit':
            return False

        # User must have permission to add A/B tests
        if not self.check_user_permissions(context['request'].user):
            return False

        return True

    @staticmethod
    def check_user_permissions(user):
        return user.has_perm('wagtail_ab_testing.add_abtest')


@hooks.register('register_page_action_menu_item')
def register_create_abtest_action_menu_item():
    return CreateAbTestActionMenuItem(order=100)


# This is the only way to inject custom JS into the editor with knowledge of the page being edited
class AbTestingTabActionMenuItem(ActionMenuItem):

    def render_html(self, context):
        if 'page' in context:
            return self.format_html(context['request'].user, context)

        return ''

    @staticmethod
    def format_html(user, context):
        return format_html(
            '<script src="{}"></script><script src="{}"></script><script>window.abTestingTabProps = JSON.parse("{}");</script>',
            reverse('wagtail_ab_testing_admin:javascript_catalog'),
            versioned_static('wagtail_ab_testing/js/wagtail-ab-testing.js'),
            escapejs(json.dumps({
                'tests': [
                    {
                        'id': ab_test.id,
                        'name': ab_test.name,
                        'started_at': ab_test.first_started_at.strftime(DATE_FORMAT) if ab_test.first_started_at else _("Not started"),
                        'status': ab_test.get_status_description(),
                        'results_url': reverse('wagtail_ab_testing_admin:results', args=[ab_test.page_id, ab_test.id]),
                    }
                    for ab_test in AbTest.objects.filter(page=context['page']).order_by('-id')
                ],
                'can_create_abtest': user.has_perm('wagtail_ab_testing.add_abtest'),
            }))
        )


@hooks.register('register_page_action_menu_item')
def register_ab_testing_tab_action_menu_item():
    return AbTestingTabActionMenuItem()


@hooks.register('after_edit_page')
def redirect_to_create_ab_test(request, page):
    if 'create-ab-test' in request.POST:
        return redirect('wagtail_ab_testing_admin:add_ab_test_compare', page.id)


@hooks.register('before_edit_page')
def check_for_running_ab_test(request, page):
    running_experiment = AbTest.objects.get_current_for_page(page=page)
    if running_experiment:
        return views.progress(request, page, running_experiment)


@hooks.register('before_serve_page')
def before_serve_page(page, request, serve_args, serve_kwargs):
    # Check if the user is trackable
    if not request_is_trackable(request):
        return

    # Check for a running A/B test on the requested page
    try:
        test = AbTest.objects.get(page=page, status=AbTest.STATUS_RUNNING)
    except AbTest.DoesNotExist:
        return

    # Save reference to test on request object so it can be found by the {% wagtail_ab_testing_script %} template tag
    request.wagtail_ab_testing_test = test

    # If this request is coming from a frontend worker, return both the control and variant versions
    # The worker will decide which version to serve to the user
    if request.META.get('HTTP_X_REQUESTED_WITH') == 'WagtailAbTestingWorker':
        if request.META.get('HTTP_AUTHORIZATION', '') != 'Token ' + settings.WAGTAIL_AB_TESTING_WORKER_TOKEN:
            raise PermissionDenied

        control_response = page.serve(request, *serve_args, **serve_kwargs)

        # Note: we must render the control response before setting `wagtail_ab_testing_serving_variant`
        if hasattr(control_response, "render"):
            control_response.render()

        request.wagtail_ab_testing_serving_variant = True

        variant_response = test.variant_revision.as_object().serve(request, *serve_args, **serve_kwargs)

        if hasattr(variant_response, "render"):
            variant_response.render()

        response = JsonResponse({
            'control': control_response.content.decode('utf-8'),
            'variant': variant_response.content.decode('utf-8'),
        })

        response['X-WagtailAbTesting-Test'] = str(test.id)

        return response

    # If the user visiting is a participant, show them the same version they saw before
    if f'wagtail-ab-testing_{test.id}_version' in request.COOKIES:
        version = request.COOKIES[f'wagtail-ab-testing_{test.id}_version']
    else:
        # Otherwise, show them the version of the page that the next participant should see.
        # Note: In order to exclude bots, the browser must call a JavaScript API to sign up as a participant
        # Once they've signed up, they'll get a cookie which keeps them on the same version
        version = test.get_new_participant_version()

    # If the user should be shown the variant, serve that from the revision. Otherwise return to keep the control
    if version == AbTest.VERSION_VARIANT:
        request.wagtail_ab_testing_serving_variant = True
        return test.variant_revision.as_object().serve(request, *serve_args, **serve_kwargs)


class AbTestingReportMenuItem(MenuItem):

    def is_shown(self, request):
        return True


@hooks.register('register_reports_menu_item')
def register_ab_testing_report_menu_item():
    return AbTestingReportMenuItem(_('A/B testing'), reverse('wagtail_ab_testing_admin:report'), icon_name='people-arrows', order=1000)


@hooks.register('register_icons')
def register_icons(icons):
    icons.append('wagtail_ab_testing/icons/people-arrows.svg')
    icons.append('wagtail_ab_testing/icons/crown.svg')
    return icons


@hooks.register('register_permissions')
def register_add_abtest_permission():
    return Permission.objects.filter(content_type__app_label='wagtail_ab_testing', codename='add_abtest')



# Generated by Django 3.1.3 on 2020-12-16 13:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_ab_testing', '0009_rename_variant_to_version'),
    ]

    operations = [
        migrations.RenameField(
            model_name='abtest',
            old_name='treatment_revision',
            new_name='variant_revision',
        ),
        migrations.AlterField(
            model_name='abtest',
            name='winning_version',
            field=models.CharField(choices=[('control', 'Control'), ('variant', 'Variant')], max_length=9, null=True),
        ),
        migrations.AlterField(
            model_name='abtesthourlylog',
            name='version',
            field=models.CharField(choices=[('control', 'Control'), ('variant', 'Variant')], max_length=9),
        ),
    ]
